require 'active_support/core_ext/string'
require 'active_support/core_ext/numeric/time'

module BackRun
  class Job
    # This is a Google Pubsub restriction. The ack deadline of a message can not be modified more
    # than 600 seconds
    MAX_REMAINING_SECONDS = 600
    MINUTES_PER_RETRY = 5.minutes
    attr_reader :queue_name, :klass, :duration_seconds

    def initialize(klass, json_args, queue_name, scheduled_at, retries)
      @klass = klass
      @args = json_args
      @queue_name = queue_name
      @scheduled_at = scheduled_at
      @retries = retries
    end

    def self.new_from_active_job(job)
      new(job.class.to_s, job.arguments.to_json, job.queue_name, job.scheduled_at, 0)
    end

    def self.from_json(json)
      attributes = JSON.parse(json)
      new(
        attributes['class'], attributes['args'], attributes['queue_name'],
        attributes['scheduled_at'], attributes['retries']
      )
    end

    def to_json(*_args)
      {
        class: @klass, args: @args, queue_name: @queue_name, scheduled_at: @scheduled_at.to_f,
        retries: @retries
      }.to_json
    end

    def perform(worker)
      register_duration { @klass.constantize.new.perform(*JSON.parse(@args)) }
      BackRun.logger.info("Finished processing the job: #{@klass}")
    rescue StandardError => e
      BackRun.logger.error("Failed processing the job: #{e.message}")
      handle_job_failure(worker)
    end

    def should_run_now?
      @scheduled_at.nil? || @scheduled_at <= Time.now.to_f
    end

    def remaining_seconds_to_run
      [((Time.at(@scheduled_at) - Time.now) / 1.second).round, MAX_REMAINING_SECONDS].min
    end

    private

    def handle_job_failure(worker)
      @retries += 1
      if @retries <= 2
        @scheduled_at = (Time.now + MINUTES_PER_RETRY).to_f
        worker.retry_job(self)
      else
        worker.kill_job(self)
      end
    end

    def register_duration
      start_time = Time.now
      yield
      end_time = Time.now
      @duration_seconds = (end_time - start_time) / 1.second
    end
  end
end
