require 'google/cloud/pubsub'
require_relative 'job'

module BackRun
  class Pubsub
    def initialize
      @pubsub = Google::Cloud::Pubsub.new(
        project_id: Rails.application.config.back_run[:project_id],
        credentials: Rails.application.config.back_run[:credentials_path]
      )
    end

    def publish(job, topic = nil)
      topic_name = topic || job.queue_name
      topic = fetch_topic(topic_name)
      topic.publish(job.to_json)
    end

    # The gem is using Concurrent::CachedThreadPool which overrides the max_threads, so it's
    # ignoring the stream configuration used in the listen method. It may cause that lot's of
    # threads are created when the traffic increases.
    # This PR https://github.com/googleapis/google-cloud-ruby/pull/3682 fixes that but it's not
    # released yet.
    def subscribe(worker)
      worker.queues.each do |queue|
        subscription = subscription_for(queue)
        subscriber = subscription.listen do |message|
          job = Job.from_json(message.data)
          worker.message_received(job, ack_callback(job, message), modify_ack_callback(message))
        end
        subscriber.on_error { |error| BackRun.logger.error(error) }
        subscriber.start
      end
    end

    def kill_job(job)
      BackRun.logger.info("Moving to morgue queue #{job.klass}")
      publish(job, 'morgue')
    end

    private

    def subscription_for(queue)
      topic = fetch_topic(queue)
      topic.subscription("worker_#{queue}") || topic.subscribe("worker_#{queue}")
    end

    def ack_callback(job, message)
      proc do
        BackRun.logger.info("Message received from queue: #{job.queue_name}")
        message.ack!
      end
    end

    def modify_ack_callback(message)
      proc do |seconds|
        BackRun.logger.info("Running job in #{seconds} seconds")
        message.modify_ack_deadline!(seconds)
      end
    end

    def fetch_topic(topic_name)
      @pubsub.topic(topic_name) || @pubsub.create_topic(topic_name)
    end
  end
end
